@@ -25,7 +25,7 @@ use super::{
 const PRIMARY_INDEX: Label = label("primary");
 
 #[derive(Clone)]
-pub struct Index<F: File<Node>, D: Dir, Txn: Transaction<D>> {
+pub struct Index<F, D, Txn> {
     btree: BTreeFile<F, D, Txn>,
     schema: IndexSchema,
 }
@@ -88,13 +88,15 @@ impl<F: File<Node>, D: Dir, Txn: Transaction<D>> Index<F, D, Txn> {
         reverse: bool,
     ) -> TCResult<TCTryStream<'a, Vec<Value>>> {
         self.validate_bounds(&bounds)?;
-
         let range = bounds.into_btree_range(&self.schema.columns())?;
         self.btree.slice(range, reverse)?.keys(txn_id).await
     }
 }
 
-impl<F: File<Node>, D: Dir, Txn: Transaction<D>> Instance for Index<F, D, Txn> {
+impl<F, D, Txn> Instance for Index<F, D, Txn>
+where
+    Self: Send + Sync,
+{
     type Class = TableType;
 
     fn class(&self) -> TableType {
@@ -256,14 +258,14 @@ where
     }
 }
 
-impl<F: File<Node>, D: Dir, Txn: Transaction<D>> From<Index<F, D, Txn>> for Table<F, D, Txn> {
+impl<F, D, Txn> From<Index<F, D, Txn>> for Table<F, D, Txn> {
     fn from(index: Index<F, D, Txn>) -> Self {
         Table::Index(index)
     }
 }
 
 #[derive(Clone)]
-pub struct ReadOnly<F: File<Node>, D: Dir, Txn: Transaction<D>> {
+pub struct ReadOnly<F, D, Txn> {
     index: IndexSlice<F, D, Txn>,
 }
 
@@ -389,7 +391,7 @@ impl<F: File<Node>, D: Dir, Txn: Transaction<D>> From<ReadOnly<F, D, Txn>> for T
     }
 }
 
-struct Inner<F: File<Node>, D: Dir, Txn: Transaction<D>> {
+struct Inner<F, D, Txn> {
     schema: TableSchema,
     primary: Index<F, D, Txn>,
     auxiliary: Vec<(Id, Index<F, D, Txn>)>,
@@ -397,14 +399,15 @@ struct Inner<F: File<Node>, D: Dir, Txn: Transaction<D>> {
 
 /// The base type of a [`Table`].
 #[derive(Clone)]
-pub struct TableIndex<F: File<Node>, D: Dir, Txn: Transaction<D>> {
+pub struct TableIndex<F, D, Txn> {
     inner: Arc<Inner<F, D, Txn>>,
 }
 
 impl<F: File<Node>, D: Dir, Txn: Transaction<D>> TableIndex<F, D, Txn> {
+    /// Create a new `TableIndex` with the given [`TableSchema`].
     pub async fn create(
-        schema: TableSchema,
         context: &D,
+        schema: TableSchema,
         txn_id: TxnId,
     ) -> TCResult<TableIndex<F, D, Txn>>
     where
@@ -464,10 +467,14 @@ impl<F: File<Node>, D: Dir, Txn: Transaction<D>> TableIndex<F, D, Txn> {
         Ok(Index { btree, schema })
     }
 
+    /// Return `true` if this table has zero rows.
     pub async fn is_empty(&self, txn: &Txn) -> TCResult<bool> {
         self.inner.primary.is_empty(txn).await
     }
 
+    /// Merge the given list of `Bounds` into a single `Bounds` instance.
+    ///
+    /// Returns an error in the case that later [`Bounds`] are larger than earlier [`Bounds`].
     pub fn merge_bounds(&self, all_bounds: Vec<Bounds>) -> TCResult<Bounds> {
         let collator = self.inner.primary.btree().collator();
 
@@ -479,10 +486,12 @@ impl<F: File<Node>, D: Dir, Txn: Transaction<D>> TableIndex<F, D, Txn> {
         Ok(merged)
     }
 
+    /// Borrow the primary `Index` of this `TableIndex`.
     pub fn primary(&self) -> &Index<F, D, Txn> {
         &self.inner.primary
     }
 
+    /// Return an index which supports the given [`Bounds`], or an error if there is none.
     pub fn supporting_index(&self, bounds: &Bounds) -> TCResult<Index<F, D, Txn>> {
         if self.inner.primary.validate_bounds(bounds).is_ok() {
             return Ok(self.inner.primary.clone());
@@ -500,10 +509,12 @@ impl<F: File<Node>, D: Dir, Txn: Transaction<D>> TableIndex<F, D, Txn> {
         ))
     }
 
+    /// Return a single row in this table with the given primary `key`, or `None` if there is none.
     pub async fn get(&self, txn_id: TxnId, key: Vec<Value>) -> TCResult<Option<Vec<Value>>> {
         self.inner.primary.get(txn_id, key).await
     }
 
+    /// Insert a new row into this `TableIndex`, or update the row at the given `key` with `values`.
     pub async fn upsert(&self, txn_id: TxnId, key: Vec<Value>, values: Vec<Value>) -> TCResult<()> {
         let primary = &self.inner.primary;
 
@@ -525,6 +536,7 @@ impl<F: File<Node>, D: Dir, Txn: Transaction<D>> TableIndex<F, D, Txn> {
         Ok(())
     }
 
+    /// Stream the rows within the given [`Bounds`] from the primary index of this `TableIndex`.
     pub async fn slice_rows<'a>(
         self,
         txn_id: TxnId,
@@ -602,92 +614,42 @@ impl<F: File<Node>, D: Dir, Txn: Transaction<D>> TableInstance<F, D, Txn>
         self.validate_order(&columns)?;
 
         let selection = TableSlice::new(self.clone(), Bounds::default())?;
-        let mut merge_source = MergeSource::Table(selection);
+        let merge_source = MergeSource::Table(selection);
 
-        let mut columns = &columns[..];
-        loop {
-            let initial = columns.to_vec();
-
-            let mut i = 0;
-            while i < columns.len() {
-                let subset = &columns[i..];
-                let mut supported = false;
-
-                if self.inner.primary.validate_order(subset).is_ok() {
-                    supported = true;
-                    columns = &columns[..i];
-                    i = 0;
+        if self.primary().validate_order(&columns).is_ok() {
+            debug!("primary key can order by {}", Tuple::from(columns.clone()));
 
+            let index_slice = self.primary().clone().index_slice(Bounds::default())?;
+            let merged = Merged::new(merge_source, index_slice)?;
+            return if reverse {
+                merged.reversed()
+            } else {
+                Ok(merged.into())
+            };
+        } else {
+            for (name, index) in &self.inner.auxiliary {
+                if index.validate_order(&columns).is_ok() {
                     debug!(
-                        "primary key supports order {}",
-                        Value::from_iter(subset.to_vec())
+                        "index {} can order by {}",
+                        name,
+                        Tuple::from(columns.clone())
                     );
 
-                    let index_slice = self.inner.primary.clone().index_slice(Bounds::default())?;
+                    let index_slice = index.clone().index_slice(Bounds::default())?;
                     let merged = Merged::new(merge_source, index_slice)?;
-
-                    if columns.is_empty() {
-                        return if reverse {
-                            merged.reversed()
-                        } else {
-                            Ok(merged.into())
-                        };
-                    }
-
-                    merge_source = MergeSource::Merge(Box::new(merged));
-                } else {
-                    debug!(
-                        "primary key does not support order {}",
-                        Value::from_iter(subset.to_vec())
-                    );
-
-                    for (name, index) in self.inner.auxiliary.iter() {
-                        if index.validate_order(subset).is_ok() {
-                            supported = true;
-                            columns = &columns[..i];
-                            i = 0;
-
-                            debug!(
-                                "index {} supports order {}",
-                                name,
-                                Value::from_iter(subset.to_vec())
-                            );
-
-                            let index_slice = index.clone().index_slice(Bounds::default())?;
-                            let merged = Merged::new(merge_source, index_slice)?;
-
-                            if columns.is_empty() {
-                                return if reverse {
-                                    merged.reversed()
-                                } else {
-                                    Ok(merged.into())
-                                };
-                            }
-
-                            merge_source = MergeSource::Merge(Box::new(merged));
-                            break;
-                        } else {
-                            debug!(
-                                "index {} does not support order {}",
-                                name,
-                                Value::from_iter(subset.to_vec())
-                            );
-                        }
-                    }
+                    return if reverse {
+                        merged.reversed()
+                    } else {
+                        Ok(merged.into())
+                    };
                 }
-
-                if !supported {
-                    i = i + 1;
-                }
-            }
-
-            if columns == &initial[..] {
-                return Err(TCError::bad_request(
-                    "This table has no index to support the order",
-                    Value::from_iter(columns.to_vec()),
-                ));
             }
         }
+
+        Err(TCError::bad_request(
+            "table has no index to order by",
+            Tuple::<Id>::from_iter(columns),
+        ))
     }
 
     fn reversed(self) -> TCResult<Self::Reverse> {
@@ -737,13 +699,17 @@ impl<F: File<Node>, D: Dir, Txn: Transaction<D>> TableInstance<F, D, Txn>
                     }
 
                     merge_source = MergeSource::Merge(Box::new(merged));
+                    break;
                 } else {
+                    let mut supported = false;
                     for (name, index) in self.inner.auxiliary.iter() {
                         debug!("checking index {} with schema {}", name, index.schema());
 
                         match index.validate_bounds(&subset) {
                             Ok(()) => {
                                 debug!("index {} can slice {}", name, subset);
+                                supported = true;
+
                                 let index_slice = index.clone().index_slice(subset)?;
                                 let merged = Merged::new(merge_source, index_slice)?;
 
@@ -760,6 +726,10 @@ impl<F: File<Node>, D: Dir, Txn: Transaction<D>> TableInstance<F, D, Txn>
                             }
                         }
                     }
+
+                    if supported {
+                        break;
+                    }
                 };
 
                 i = i - 1;
@@ -1028,7 +998,7 @@ where
         let txn_id = *txn.id();
         let schema = source.schema();
         let key_len = schema.primary().key().len();
-        let table = Self::create(schema, &dir, txn_id).await?;
+        let table = Self::create(&dir, schema, txn_id).await?;
 
         let rows = source.rows(txn_id).await?;
 
