@@ -28,13 +28,14 @@ const ERR_LIMITED_REVERSE: &str = "Cannot reverse a limited selection. \
 Consider reversing a slice before limiting";
 
 #[derive(Clone)]
-pub struct Aggregate<F: File<Node>, D: Dir, Txn: Transaction<D>, T: TableInstance<F, D, Txn>> {
+pub struct Aggregate<F, D, Txn, T> {
     source: Selection<F, D, Txn, T>,
-    phantom_file: PhantomData<F>,
+    file: PhantomData<F>,
 }
 
-impl<F: File<Node>, D: Dir, Txn: Transaction<D>, T: TableInstance<F, D, Txn>> Instance
-    for Aggregate<F, D, Txn, T>
+impl<F, D, Txn, T> Instance for Aggregate<F, D, Txn, T>
+where
+    Self: Send + Sync,
 {
     type Class = TableType;
 
@@ -71,15 +72,15 @@ impl<F: File<Node>, D: Dir, Txn: Transaction<D>, T: TableInstance<F, D, Txn>>
         let source = self.source.order_by(columns, reverse)?;
         Ok(Aggregate {
             source,
-            phantom_file: PhantomData,
+            file: PhantomData,
         })
     }
 
     fn reversed(self) -> TCResult<Self::Reverse> {
-        let phantom_file = self.phantom_file;
+        let phantom_file = self.file;
         self.source.reversed().map(|source| Aggregate {
             source,
-            phantom_file,
+            file: phantom_file,
         })
     }
 
@@ -107,18 +108,18 @@ impl<F: File<Node>, D: Dir, Txn: Transaction<D>, T: TableInstance<F, D, Txn>>
             schema: aggregate.source.schema,
             columns: aggregate.source.columns,
             indices: aggregate.source.indices,
-            phantom: PhantomFDT::default(),
+            phantom: Phantom::default(),
         };
 
         Table::Aggregate(Box::new(Aggregate {
             source,
-            phantom_file: aggregate.phantom_file,
+            file: aggregate.file,
         }))
     }
 }
 
 #[derive(Clone)]
-pub struct IndexSlice<F: File<Node>, D: Dir, Txn: Transaction<D>> {
+pub struct IndexSlice<F, D, Txn> {
     source: BTreeFile<F, D, Txn>,
     schema: IndexSchema,
     bounds: Bounds,
@@ -212,7 +213,10 @@ impl<F: File<Node>, D: Dir, Txn: Transaction<D>> IndexSlice<F, D, Txn> {
     }
 }
 
-impl<F: File<Node>, D: Dir, Txn: Transaction<D>> Instance for IndexSlice<F, D, Txn> {
+impl<F, D, Txn> Instance for IndexSlice<F, D, Txn>
+where
+    Self: Send + Sync,
+{
     type Class = TableType;
 
     fn class(&self) -> Self::Class {
@@ -221,8 +225,11 @@ impl<F: File<Node>, D: Dir, Txn: Transaction<D>> Instance for IndexSlice<F, D, T
 }
 
 #[async_trait]
-impl<F: File<Node>, D: Dir, Txn: Transaction<D>> TableInstance<F, D, Txn>
-    for IndexSlice<F, D, Txn>
+impl<F, D, Txn> TableInstance<F, D, Txn> for IndexSlice<F, D, Txn>
+where
+    F: File<Node>,
+    D: Dir,
+    Txn: Transaction<D>,
 {
     type OrderBy = Self;
     type Reverse = Self;
@@ -307,25 +314,28 @@ impl<F: File<Node>, D: Dir, Txn: Transaction<D>> TableInstance<F, D, Txn>
     }
 }
 
-impl<F: File<Node>, D: Dir, Txn: Transaction<D>> From<IndexSlice<F, D, Txn>> for Table<F, D, Txn> {
+impl<F, D, Txn> From<IndexSlice<F, D, Txn>> for Table<F, D, Txn> {
     fn from(slice: IndexSlice<F, D, Txn>) -> Self {
         Self::IndexSlice(slice)
     }
 }
 
 #[derive(Clone)]
-pub struct Limited<F: File<Node>, D: Dir, Txn: Transaction<D>> {
+pub struct Limited<F, D, Txn> {
     source: Table<F, D, Txn>,
     limit: u64,
 }
 
-impl<F: File<Node>, D: Dir, Txn: Transaction<D>> Limited<F, D, Txn> {
+impl<F, D, Txn> Limited<F, D, Txn> {
     pub fn new(source: Table<F, D, Txn>, limit: u64) -> Self {
         Limited { source, limit }
     }
 }
 
-impl<F: File<Node>, D: Dir, Txn: Transaction<D>> Instance for Limited<F, D, Txn> {
+impl<F, D, Txn> Instance for Limited<F, D, Txn>
+where
+    Self: Send + Sync,
+{
     type Class = TableType;
 
     fn class(&self) -> Self::Class {
@@ -407,14 +417,14 @@ impl<F: File<Node>, D: Dir, Txn: Transaction<D>> TableInstance<F, D, Txn> for Li
     }
 }
 
-impl<F: File<Node>, D: Dir, Txn: Transaction<D>> From<Limited<F, D, Txn>> for Table<F, D, Txn> {
+impl<F, D, Txn> From<Limited<F, D, Txn>> for Table<F, D, Txn> {
     fn from(limited: Limited<F, D, Txn>) -> Self {
         Table::Limit(Box::new(limited))
     }
 }
 
 #[derive(Clone)]
-pub enum MergeSource<F: File<Node>, D: Dir, Txn: Transaction<D>> {
+pub enum MergeSource<F, D, Txn> {
     Table(TableSlice<F, D, Txn>),
     Merge(Box<Merged<F, D, Txn>>),
 }
@@ -434,14 +444,14 @@ impl<F: File<Node>, D: Dir, Txn: Transaction<D>> MergeSource<F, D, Txn> {
         }
     }
 
-    fn into_reversed(self) -> MergeSource<F, D, Txn> {
+    fn reversed(self) -> TCResult<MergeSource<F, D, Txn>> {
         match self {
-            Self::Table(table_slice) => Self::Table(table_slice.into_reversed()),
-            Self::Merge(merged) => Self::Merge(Box::new(merged.into_reversed())),
+            Self::Table(table_slice) => table_slice.reversed().map(Self::Table),
+            Self::Merge(merged) => merged.reversed().map(Box::new).map(Self::Merge),
         }
     }
 
-    pub async fn slice_rows<'a>(
+    async fn slice_rows<'a>(
         self,
         txn_id: TxnId,
         bounds: Bounds,
@@ -475,68 +485,26 @@ impl<F: File<Node>, D: Dir, Txn: Transaction<D>> MergeSource<F, D, Txn> {
     }
 }
 
+/// A merge of multiple table indexes
 #[derive(Clone)]
-pub struct Merged<F: File<Node>, D: Dir, Txn: Transaction<D>> {
-    key_columns: Vec<Column>,
+pub struct Merged<F, D, Txn> {
     left: MergeSource<F, D, Txn>,
     right: IndexSlice<F, D, Txn>,
     bounds: Bounds,
-    keys: Selection<F, D, Txn, IndexSlice<F, D, Txn>>,
 }
 
 impl<F: File<Node>, D: Dir, Txn: Transaction<D>> Merged<F, D, Txn> {
+    /// Create a new merge of the given `IndexSlice` with the given `MergeSource`.
     pub fn new(left: MergeSource<F, D, Txn>, right: IndexSlice<F, D, Txn>) -> TCResult<Self> {
-        let key_columns = left.key().to_vec();
-        let key_names: Vec<Id> = key_columns.iter().map(|c| c.name()).cloned().collect();
-        let keys = right.clone().select(key_names)?;
-
-        left.source()
-            .merge_bounds(vec![left.bounds().clone(), right.bounds().clone()])
-            .map(|bounds| Merged {
-                key_columns,
-                left,
-                right,
-                bounds,
-                keys,
-            })
-    }
-
-    pub async fn slice_rows<'a>(
-        self,
-        txn_id: TxnId,
-        bounds: Bounds,
-        reverse: bool,
-    ) -> TCResult<TCTryStream<'a, Vec<Value>>> {
-        let bounds = self
+        let bounds = left
             .source()
-            .merge_bounds(vec![self.bounds.clone(), bounds])?;
+            .merge_bounds(vec![left.bounds().clone(), right.bounds().clone()])?;
 
-        self.into_source().slice_rows(txn_id, bounds, reverse).await
-    }
-
-    fn into_reversed(self) -> Self {
-        let key_names = self
-            .key_columns
-            .iter()
-            .map(|col| col.name())
-            .cloned()
-            .collect();
-
-        let keys = Selection {
-            source: self.right.clone().into_reversed(),
-            schema: self.keys.schema.clone(),
-            columns: key_names,
-            indices: self.keys.indices.clone(),
-            phantom: PhantomFDT::default(),
-        };
-
-        Merged {
-            key_columns: self.key_columns.to_vec(),
-            left: self.left.into_reversed(),
-            right: self.right.into_reversed(),
-            bounds: self.bounds.clone(),
-            keys,
-        }
+        Ok(Self {
+            left,
+            right,
+            bounds,
+        })
     }
 
     fn source(&'_ self) -> &'_ TableIndex<F, D, Txn> {
@@ -546,9 +514,25 @@ impl<F: File<Node>, D: Dir, Txn: Transaction<D>> Merged<F, D, Txn> {
     fn into_source(self) -> TableIndex<F, D, Txn> {
         self.left.into_source()
     }
+
+    /// Stream the rows within the given [`Bounds`] of this merge
+    pub async fn slice_rows<'a>(
+        self,
+        txn_id: TxnId,
+        bounds: Bounds,
+        reverse: bool,
+    ) -> TCResult<TCTryStream<'a, Vec<Value>>> {
+        let these_bounds = self.bounds;
+        let source = self.left.into_source();
+        let bounds = source.merge_bounds(vec![these_bounds, bounds])?;
+        source.slice_rows(txn_id, bounds, reverse).await
+    }
 }
 
-impl<F: File<Node>, D: Dir, Txn: Transaction<D>> Instance for Merged<F, D, Txn> {
+impl<F, D, Txn> Instance for Merged<F, D, Txn>
+where
+    Self: Send + Sync,
+{
     type Class = TableType;
 
     fn class(&self) -> Self::Class {
@@ -569,7 +553,7 @@ impl<F: File<Node>, D: Dir, Txn: Transaction<D>> TableInstance<F, D, Txn> for Me
 
         rows.map(|row| row.and_then(|row| schema.row_from_values(row)))
             .map_ok(|row| self.delete_row(txn_id, row))
-            .try_buffer_unordered(2)
+            .try_buffer_unordered(num_cpus::get())
             .try_fold((), |_, _| future::ready(Ok(())))
             .await
     }
@@ -607,31 +591,33 @@ impl<F: File<Node>, D: Dir, Txn: Transaction<D>> TableInstance<F, D, Txn> for Me
     }
 
     fn reversed(self) -> TCResult<Self::Reverse> {
-        Ok(self.into_reversed())
+        Ok(Merged {
+            left: self.left.reversed()?,
+            right: self.right.reversed()?,
+            bounds: self.bounds,
+        })
     }
 
     fn slice(self, bounds: Bounds) -> TCResult<Self::Slice> {
-        let bounds = self
-            .source()
-            .merge_bounds(vec![self.bounds.clone(), bounds])?;
-
-        self.left.into_source().slice(bounds)
+        let source = self.left.into_source();
+        let bounds = source.merge_bounds(vec![self.bounds, bounds])?;
+        source.slice(bounds)
     }
 
     async fn rows<'a>(self, txn_id: TxnId) -> TCResult<TCTryStream<'a, Vec<Value>>> {
+        let key_columns = self.key().to_vec();
+        let key_names = key_columns.iter().map(|col| &col.name).cloned().collect();
+        let keys = self.right.select(key_names)?.rows(txn_id).await?;
+
         let left = self.left;
         let left_clone = left.clone();
-        let key_columns = self.key_columns;
-        let keys = self.keys.clone().rows(txn_id).await?;
-
-        let rows = keys
+        let merge = keys
             .map_ok(move |key| Bounds::from_key(key, &key_columns))
             .try_filter(move |bounds| future::ready(left.validate_bounds(bounds).is_ok()))
             .and_then(move |bounds| Box::pin(left_clone.clone().slice_rows(txn_id, bounds, false)))
             .try_flatten();
 
-        let rows: TCTryStream<Vec<Value>> = Box::pin(rows);
-        Ok(rows)
+        Ok(Box::pin(merge))
     }
 
     fn validate_bounds(&self, bounds: &Bounds) -> TCResult<()> {
@@ -669,19 +655,19 @@ impl<F: File<Node>, D: Dir, Txn: Transaction<D>> TableInstance<F, D, Txn> for Me
     }
 }
 
-impl<F: File<Node>, D: Dir, Txn: Transaction<D>> From<Merged<F, D, Txn>> for Table<F, D, Txn> {
+impl<F, D, Txn> From<Merged<F, D, Txn>> for Table<F, D, Txn> {
     fn from(merged: Merged<F, D, Txn>) -> Self {
         Self::Merge(merged)
     }
 }
 
 #[derive(Clone)]
-pub struct Selection<F: File<Node>, D: Dir, Txn: Transaction<D>, T: TableInstance<F, D, Txn>> {
+pub struct Selection<F, D, Txn, T> {
     source: T,
     schema: IndexSchema,
     columns: Vec<Id>,
     indices: Vec<usize>,
-    phantom: PhantomFDT<F, D, Txn>,
+    phantom: Phantom<F, D, Txn>,
 }
 
 impl<F: File<Node>, D: Dir, Txn: Transaction<D>, T: TableInstance<F, D, Txn>>
@@ -697,13 +683,12 @@ impl<F: File<Node>, D: Dir, Txn: Transaction<D>, T: TableInstance<F, D, Txn>>
         }
 
         let mut indices: Vec<usize> = Vec::with_capacity(columns.len());
-        let mut schema: Vec<Column> = Vec::with_capacity(columns.len());
 
-        let source_columns = [source.key(), source.values()].concat();
+        let source_columns = source.schema().primary().columns();
         let source_indices: HashMap<&Id, usize> = source_columns
             .iter()
             .enumerate()
-            .map(|(i, col)| (col.name(), i))
+            .map(|(i, col)| (&col.name, i))
             .collect();
 
         for name in columns.iter() {
@@ -712,22 +697,37 @@ impl<F: File<Node>, D: Dir, Txn: Transaction<D>, T: TableInstance<F, D, Txn>>
                 .ok_or(TCError::not_found(format!("Column {}", name)))?;
 
             indices.push(index);
-            schema.push(source_columns[index].clone());
         }
 
-        let schema = (vec![], schema).into();
+        let key = source
+            .key()
+            .iter()
+            .filter(|col| column_set.contains(&col.name))
+            .cloned()
+            .collect();
+
+        let values = source
+            .values()
+            .iter()
+            .filter(|col| column_set.contains(&col.name))
+            .cloned()
+            .collect();
+
+        let schema = (key, values).into();
+
         Ok(Selection {
             source,
             schema,
             columns,
             indices,
-            phantom: PhantomFDT::default(),
+            phantom: Phantom::default(),
         })
     }
 }
 
-impl<F: File<Node>, D: Dir, Txn: Transaction<D>, T: TableInstance<F, D, Txn>> Instance
-    for Selection<F, D, Txn, T>
+impl<F, D, Txn, T> Instance for Selection<F, D, Txn, T>
+where
+    Self: Send + Sync,
 {
     type Class = TableType;
 
@@ -737,8 +737,12 @@ impl<F: File<Node>, D: Dir, Txn: Transaction<D>, T: TableInstance<F, D, Txn>> In
 }
 
 #[async_trait]
-impl<F: File<Node>, D: Dir, Txn: Transaction<D>, T: TableInstance<F, D, Txn>>
-    TableInstance<F, D, Txn> for Selection<F, D, Txn, T>
+impl<F, D, Txn, T> TableInstance<F, D, Txn> for Selection<F, D, Txn, T>
+where
+    F: File<Node>,
+    D: Dir,
+    Txn: Transaction<D>,
+    T: TableInstance<F, D, Txn>,
 {
     type OrderBy = Selection<F, D, Txn, <T as TableInstance<F, D, Txn>>::OrderBy>;
     type Reverse = Selection<F, D, Txn, <T as TableInstance<F, D, Txn>>::Reverse>;
@@ -783,7 +787,7 @@ impl<F: File<Node>, D: Dir, Txn: Transaction<D>, T: TableInstance<F, D, Txn>>
             schema: self.schema,
             columns: self.columns,
             indices: self.indices,
-            phantom: PhantomFDT::default(),
+            phantom: Phantom::default(),
         })
     }
 
@@ -847,8 +851,12 @@ impl<F: File<Node>, D: Dir, Txn: Transaction<D>, T: TableInstance<F, D, Txn>>
     }
 }
 
-impl<F: File<Node>, D: Dir, Txn: Transaction<D>, T: TableInstance<F, D, Txn>>
-    From<Selection<F, D, Txn, T>> for Table<F, D, Txn>
+impl<F, D, Txn, T> From<Selection<F, D, Txn, T>> for Table<F, D, Txn>
+where
+    F: File<Node>,
+    D: Dir,
+    Txn: Transaction<D>,
+    T: TableInstance<F, D, Txn>,
 {
     fn from(selection: Selection<F, D, Txn, T>) -> Self {
         Table::Selection(Box::new(Selection {
@@ -856,13 +864,13 @@ impl<F: File<Node>, D: Dir, Txn: Transaction<D>, T: TableInstance<F, D, Txn>>
             schema: selection.schema,
             columns: selection.columns,
             indices: selection.indices,
-            phantom: PhantomFDT::default(),
+            phantom: Phantom::default(),
         }))
     }
 }
 
 #[derive(Clone)]
-pub struct TableSlice<F: File<Node>, D: Dir, Txn: Transaction<D>> {
+pub struct TableSlice<F, D, Txn> {
     table: TableIndex<F, D, Txn>,
     slice: IndexSlice<F, D, Txn>,
 }
@@ -886,14 +894,14 @@ impl<F: File<Node>, D: Dir, Txn: Transaction<D>> TableSlice<F, D, Txn> {
         self.slice.slice_index(bounds)
     }
 
-    pub fn source(&'_ self) -> &'_ TableIndex<F, D, Txn> {
-        &self.table
-    }
-
     pub fn into_source(self) -> TableIndex<F, D, Txn> {
         self.table
     }
 
+    pub fn source(&'_ self) -> &'_ TableIndex<F, D, Txn> {
+        &self.table
+    }
+
     pub async fn slice_rows<'a>(
         self,
         txn_id: TxnId,
@@ -902,16 +910,12 @@ impl<F: File<Node>, D: Dir, Txn: Transaction<D>> TableSlice<F, D, Txn> {
     ) -> TCResult<TCTryStream<'a, Vec<Value>>> {
         self.slice.slice_rows(txn_id, bounds, reverse).await
     }
-
-    fn into_reversed(self) -> Self {
-        TableSlice {
-            table: self.table,
-            slice: self.slice.into_reversed(),
-        }
-    }
 }
 
-impl<F: File<Node>, D: Dir, Txn: Transaction<D>> Instance for TableSlice<F, D, Txn> {
+impl<F, D, Txn> Instance for TableSlice<F, D, Txn>
+where
+    Self: Send + Sync,
+{
     type Class = TableType;
 
     fn class(&self) -> Self::Class {
@@ -920,8 +924,11 @@ impl<F: File<Node>, D: Dir, Txn: Transaction<D>> Instance for TableSlice<F, D, T
 }
 
 #[async_trait]
-impl<F: File<Node>, D: Dir, Txn: Transaction<D>> TableInstance<F, D, Txn>
-    for TableSlice<F, D, Txn>
+impl<F, D, Txn> TableInstance<F, D, Txn> for TableSlice<F, D, Txn>
+where
+    F: File<Node>,
+    D: Dir,
+    Txn: Transaction<D>,
 {
     type OrderBy = Merged<F, D, Txn>;
     type Reverse = TableSlice<F, D, Txn>;
@@ -973,13 +980,11 @@ impl<F: File<Node>, D: Dir, Txn: Transaction<D>> TableInstance<F, D, Txn>
     }
 
     fn slice(self, bounds: Bounds) -> TCResult<Merged<F, D, Txn>> {
-        let bounds = self
-            .source()
-            .merge_bounds(vec![self.slice.bounds().clone(), bounds])?;
-
-        self.validate_bounds(&bounds)?;
-
-        self.into_source().slice(bounds)
+        let slice_bounds = self.slice.bounds().clone();
+        let source = self.into_source();
+        let bounds = source.merge_bounds(vec![slice_bounds, bounds])?;
+        source.validate_bounds(&bounds)?;
+        source.slice(bounds)
     }
 
     async fn rows<'a>(self, txn_id: TxnId) -> TCResult<TCTryStream<'a, Vec<Value>>> {
@@ -987,8 +992,6 @@ impl<F: File<Node>, D: Dir, Txn: Transaction<D>> TableInstance<F, D, Txn>
     }
 
     fn validate_bounds(&self, bounds: &Bounds) -> TCResult<()> {
-        debug!("Table::validate_bounds {}", bounds);
-
         let index = self.source().supporting_index(self.slice.bounds())?;
         index
             .validate_slice_bounds(self.slice.bounds().clone(), bounds.clone())
@@ -1019,7 +1022,7 @@ impl<F: File<Node>, D: Dir, Txn: Transaction<D>> TableInstance<F, D, Txn>
     }
 }
 
-impl<F: File<Node>, D: Dir, Txn: Transaction<D>> From<TableSlice<F, D, Txn>> for Table<F, D, Txn> {
+impl<F, D, Txn> From<TableSlice<F, D, Txn>> for Table<F, D, Txn> {
     fn from(slice: TableSlice<F, D, Txn>) -> Self {
         Self::TableSlice(slice)
     }
@@ -1031,20 +1034,21 @@ pub fn group_by<F: File<Node>, D: Dir, Txn: Transaction<D>, T: TableInstance<F,
 ) -> TCResult<Aggregate<F, D, Txn, <T as TableInstance<F, D, Txn>>::OrderBy>> {
     let source = source.order_by(columns.to_vec(), false)?;
     let source = source.select(columns)?;
+
     Ok(Aggregate {
         source,
-        phantom_file: PhantomData,
+        file: PhantomData,
     })
 }
 
 #[derive(Clone)]
-struct PhantomFDT<F: File<Node>, D: Dir, Txn: Transaction<D>> {
+struct Phantom<F, D, Txn> {
     file: PhantomData<F>,
     dir: PhantomData<D>,
     txn: PhantomData<Txn>,
 }
 
-impl<F: File<Node>, D: Dir, Txn: Transaction<D>> Default for PhantomFDT<F, D, Txn> {
+impl<F: File<Node>, D: Dir, Txn: Transaction<D>> Default for Phantom<F, D, Txn> {
     fn default() -> Self {
         Self {
             file: PhantomData,
